"""
WildCard Matching Implementation
"""
from typing import Optional, List
from typing_extensions import Self

from pyderive import dataclass, field

#** Variables **#
__all__ = ['WildcardMatch']

#** Classes **#

@dataclass(slots=True)
class WildcardMatch:
    """
    Match a String against a Wildcarded Pattern
    """
    prefix: Optional[bytes] = None
    middle: List[bytes]     = field(default_factory=list)
    suffix: Optional[bytes] = None

    @classmethod
    def compile(cls, pattern: str) -> Self:
        """
        Compile the specified pattern into a matcher object

        :param pattern: wildcard pattern to match
        :return:        compiled pattern matcher
        """
        chunks = [c for c in pattern.encode().split(b'*')]
        prefix = chunks.pop(0)
        suffix = chunks.pop() if chunks else None
        return cls(prefix or None, chunks, suffix or None)

    def match(self, string: bytes) -> bool:
        """
        determine if specified string matches the compiled pattern

        :param string: string to compare to pattern
        :return:       true if pattern matches compiled expression
        """
        if self.prefix and not string.startswith(self.prefix):
            return False
        index = 0
        for middle in self.middle:
            sidx = string.find(middle, index)
            if sidx < 0:
                return False
            index = sidx + 1
        if self.suffix and not string.endswith(self.suffix, index):
            return False
        return True



"""
DNS Question Object Definitions
"""
from pystructs import U16, Domain, Struct
from typing_extensions import Annotated

from .enum import RType, RClass

#** Variables **#
__all__ = ['Question', 'Zone']

#** Classes **#

class Question(Struct):
    """
    DNS Question Object Definition
    """
    name:   Domain
    qtype:  Annotated[RType, U16]
    qclass: Annotated[RClass, U16] = RClass.IN

class Zone(Question):
    """
    Alias of Question in UPDATE action DNS Requests
    """
    pass



"""
DNS Message Packing/Unpacking UnitTests
"""
from unittest import TestCase

from .. import Message, OpCode, QR, RClass, RCode, RType
from ..edns import EdnsAnswer

#** Variables **#
__all__ = ['MessageTests']

EXAMPLE_REQUEST = 'a32701200001000000000001076578616d706c6503' +\
    '636f6d000001000100002904d000000000000c000a0008c5a01ecf50bd546c'

EXAMPLE_RESPONSE = 'a32781a00001000100000001076578616d706c650' +\
    '3636f6d0000010001c00c000100010000001e00045db8d70e0000290200000000000000'

#** Classes **#

class MessageTests(TestCase):
    """
    DNS Message Packet Parsing/Construction UnitTests
    """

    def test_parse_request(self):
        """
        ensure request parsing works as intended
        """
        data    = bytes.fromhex(EXAMPLE_REQUEST)
        request = Message.unpack(data)
        self.assertEqual(request.id, 0xa327)
        self.assertEqual(request.flags.qr, QR.Question)
        self.assertEqual(request.flags.op, OpCode.Query)
        self.assertEqual(request.flags.authorative, False)
        self.assertEqual(request.flags.truncated, False)
        self.assertEqual(request.flags.recursion_desired, True)
        self.assertEqual(request.flags.recursion_available, False)
        self.assertEqual(request.flags.answer_authenticated, True)
        self.assertEqual(request.flags.checking_disabled, False)
        self.assertEqual(request.flags.rcode, RCode.NoError)
        self.assertEqual(len(request.questions), 1)
        self.assertEqual(len(request.answers), 0)
        self.assertEqual(len(request.authority), 0)
        self.assertEqual(len(request.additional), 1)
        self.assertEqual(request.questions[0].name, b'example.com')
        self.assertEqual(request.questions[0].qtype, RType.A)
        self.assertEqual(request.questions[0].qclass, RClass.IN)
        self.assertIsInstance(request.additional[0], EdnsAnswer)
        self.assertEqual(request.additional[0].name, b'')

    def test_pack_request(self):
        """
        ensure request packing works as intended
        """
        data    = bytes.fromhex(EXAMPLE_REQUEST)
        request = Message.unpack(data)
        data_2  = request.pack()
        self.assertEqual(data, data_2)

    def test_parse_response(self):
        """
        ensure response parsing works as intended
        """
        data     = bytes.fromhex(EXAMPLE_RESPONSE)
        response = Message.unpack(data)
        self.assertEqual(response.id, 0xa327)
        self.assertEqual(response.flags.qr, QR.Response)
        self.assertEqual(response.flags.op, OpCode.Query)
        self.assertEqual(response.flags.authorative, False)
        self.assertEqual(response.flags.truncated, False)
        self.assertEqual(response.flags.recursion_desired, True)
        self.assertEqual(response.flags.recursion_available, True)
        self.assertEqual(response.flags.answer_authenticated, True)
        self.assertEqual(response.flags.checking_disabled, False)
        self.assertEqual(response.flags.rcode, RCode.NoError)
        self.assertEqual(len(response.questions), 1)
        self.assertEqual(len(response.answers), 1)
        self.assertEqual(len(response.authority), 0)
        self.assertEqual(len(response.additional), 1)
        self.assertEqual(response.questions[0].name, b'example.com')
        self.assertEqual(response.questions[0].qtype, RType.A)
        self.assertEqual(response.questions[0].qclass, RClass.IN)
        self.assertEqual(response.answers[0].name, b'example.com')
        self.assertEqual(response.answers[0].ttl, 30)
        self.assertEqual(response.answers[0].rtype, RType.A)
        self.assertEqual(str(response.answers[0].content.ip), '93.184.215.14') #type: ignore
        self.assertEqual(response.answers[0].rclass, RClass.IN)
        self.assertIsInstance(response.additional[0], EdnsAnswer)
        self.assertEqual(response.additional[0].name, b'')

    def test_pack_response(self):
        """
        ensure response packing works as intended
        """
        data     = bytes.fromhex(EXAMPLE_REQUEST)
        response = Message.unpack(data)
        data_2   = response.pack()
        self.assertEqual(data, data_2)



"""
Backend Extension to support In-Memory Answer Caching
"""
import time
import math
from logging import Logger, getLogger
from threading import Lock
from typing import ClassVar, Dict, List, Optional, Set

from pyderive import InitVar, dataclass, field

from . import Answers, Backend, RType, Answer
from .memory import MemoryBackend
from .ruleset import RuleBackend

#** Variables **#
__all__ = ['Cache']

#: default set of other backend sources to ignore
IGNORE = {MemoryBackend.source, RuleBackend.source}

#** Classes **#

@dataclass(slots=True)
class CacheRecord:
    """
    Record Entry for In-Memory Cache
    """
    answers:    List[Answer]
    expiration: InitVar[int]
    expires:    float = field(init=False)
    accessed:   float = field(init=False)

    def __post_init__(self, expiration: int): #type: ignore
        """
        calculate expiration-time and last-accessed time
        """
        ttl = min(a.ttl for a in self.answers)
        ttl = min(ttl, expiration) if expiration else ttl
        now = time.time()
        self.expires  = now + ttl
        self.accessed = now

    def is_expired(self) -> bool:
        """
        calculate if expiration has passed or ttl is expired
        """
        now = time.time()
        if self.expires <= now:
            return True
        elapsed = math.floor(now - self.accessed)
        if not elapsed:
            return False
        for answer in self.answers:
            answer.ttl -= elapsed
            if answer.ttl <= 0:
                return True
        self.accessed = now
        return False

@dataclass(slots=True, repr=False)
class Cache(Backend):
    """
    In-Memory Cache Extension for Backend Results
    """
    source: ClassVar[str] = 'Cache'

    backend:        Backend
    expiration:     int        = 30
    maxsize:        int        = 10000
    ignore_rtypes:  Set[RType] = field(default_factory=lambda: {RType.SOA, })
    ignore_sources: Set[str]   = field(default_factory=lambda: IGNORE)
    logger:         Logger     = field(default_factory=lambda: getLogger('pydns'))

    mutex:       Lock                   = field(default_factory=Lock, init=False)
    cache:       Dict[str, CacheRecord] = field(default_factory=dict, init=False)
    authorities: Dict[bytes, bool]      = field(default_factory=dict, init=False)

    recursion_available: bool = field(default=False, init=False)

    def __post_init__(self):
        self.logger              = self.logger.getChild('cache')
        self.recursion_available = self.backend.recursion_available

    def is_authority(self, domain: bytes) -> bool:
        """
        retrieve if domain is authority from cache before checking backend
        """
        # check cache before querying backend
        if domain in self.authorities:
            return self.authorities[domain]
        # query backend and then permanently cache authority result
        authority = self.backend.is_authority(domain)
        with self.mutex:
            if len(self.authorities) >= self.maxsize:
                self.authorities.clear()
            self.authorities[domain] = authority
        return authority

    def get_cache(self, domain: bytes, rtype: RType) -> Optional[Answers]:
        """
        retrieve from cache directly if present
        """
        key = f'{domain}->{rtype.name}'
        with self.mutex:
            if key not in self.cache:
                return
            record = self.cache[key]
            if record.is_expired():
                self.logger.debug(f'{key} expired')
                del self.cache[key]
                return
            return Answers(record.answers, self.source)

    def set_cache(self, domain: bytes, rtype: RType, answers: Answers):
        """
        save the given answers to cache for the specified domain/rtype
        """
        if not answers.answers:
            self.logger.debug(f'cannot cache empty record for {domain!r}')
            return
        key = f'{domain}->{rtype.name}'
        with self.mutex:
            if len(self.cache) >= self.maxsize:
                self.logger.debug(f'maxsize: {self.maxsize} exceeded. clearing cache!')
                self.cache.clear()
            self.cache[key] = CacheRecord(answers.answers, self.expiration)

    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        """
        retrieve answers from cache before checking supplied backend
        """
        # attempt to retrieve from cache if it exists
        answers = self.get_cache(domain, rtype)
        if answers is not None:
            return answers
        # complete standard lookup for answers
        answers = self.backend.get_answers(domain, rtype)
        if answers.source in self.ignore_sources:
            return answers
        if rtype not in self.ignore_rtypes \
            and all(a.rtype in self.ignore_rtypes for a in answers.answers):
            return answers
        # save results to cache and return results
        self.set_cache(domain, rtype, answers)
        return answers



"""
DNS Server Data Backend Implementations
"""
from abc import abstractmethod
from typing import NamedTuple, Protocol, List, ClassVar

from ... import RType, Answer

#** Variables **#
__all__ = [
    'Answers',
    'Backend',

    'Cache',
    'Forwarder',
    'MemoryBackend',

    'RuleEngine',
    'RuleBackend',
    'DbmRuleEngine',
]

#** Classes **#

class Answers(NamedTuple):
    """
    Backend DNS Answers Return Type
    """
    answers: List[Answer]
    source:  str

class Backend(Protocol):
    """
    BaseClass Interface Definition for Backend Implementations
    """
    source: ClassVar[str]
    recursion_available: bool = False

    @abstractmethod
    def is_authority(self, domain: bytes) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        raise NotImplementedError

#** Imports **#
from .cache import Cache
from .forwarder import Forwarder
from .memory import MemoryBackend
from .ruleset import RuleEngine, RuleBackend, DbmRuleEngine



"""
In-Memory Backend Implementation
"""
import ipaddress
from typing import Any, ClassVar, Dict, List, Set, Union
from typing_extensions import TypedDict

from . import Backend, Answers, Answer, RType
from ...answer import get_ctype
from ...content import Content, PTR

#** Variables **#
__all__ = ['MemoryBackend']

#: type definiton for in-memory record database
RecordDB = Dict[bytes, Dict[RType, List[Answer]]]

RecordEntries = List[Union[Content, 'RecordDict']]

RawRecordEntries = Dict[str, List[Dict[str, Any]]]

#** Classes **#

class RecordDict(TypedDict):
    content: Content
    ttl:     int

class MemoryBackend(Backend):
    """
    Simple In-Memory Backend for DNS Records
    """
    source:              ClassVar[str]  = 'MemDB'
    recursion_available: ClassVar[bool] = False #type: ignore

    __slots__ = ('records', 'authorities', 'default_ttl')

    def __init__(self, default_ttl: int = 60):
        self.records:     RecordDB   = {}
        self.authorities: Set[bytes] = set()
        self.default_ttl: int        = default_ttl

    def add_answer(self, domain: bytes, answer: Answer):
        """
        add additional domain answer into in-memory db
        """
        rtype = answer.rtype
        self.records.setdefault(domain, {})
        self.records[domain].setdefault(rtype, [])
        self.records[domain][rtype].append(answer)

    def save_domain(self, domain: bytes, records: RecordEntries):
        """
        save additional records into in-memory db

        :param domain:  dns domain to save records for
        :param records: list of record objects with ttls
        """
        assert isinstance(domain, bytes), 'domain must be bytes'
        # convert dictionary records into valid dns answer objects
        self.authorities.add(domain)
        for record in records:
            # convert record into content and ttl
            if isinstance(record, Content):
                ttl     = self.default_ttl
                content = record
            else:
                ttl     = record.get('ttl', self.default_ttl)
                content = record['content']
            # add standard record
            answer  = Answer(domain, ttl, content)
            self.add_answer(domain, answer)
            # reverse ip records for PTR lookups
            ipaddr = getattr(content, 'ip', None)
            if ipaddr is not None:
                ip   = ipaddress.ip_address(ipaddr)
                name = ip.reverse_pointer.encode()
                answer = Answer(name, ttl, PTR(domain))
                self.add_answer(name, answer)

    def save_domain_dict(self, domain: bytes, entries: RawRecordEntries):
        """
        save additional records into in-memory db using raw dictionary object

        :param domain:  dns domain to save records for
        :param records: list of record objects with ttls
        """
        records = []
        for rname, r_entries in entries.items():
            rtype  = RType[rname]
            ctype  = get_ctype(rtype)
            for entry in r_entries:
                ttl     = entry.pop('ttl', self.default_ttl)
                content = ctype(**entry)
                records.append({'ttl': ttl, 'content': content})
        self.save_domain(domain, records)

    def is_authority(self, domain: bytes) -> bool:
        """
        retrieve if domain is an authority
        """
        return domain in self.authorities

    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        """
        retrieve records associated w/ the given domain and record-type
        """
        answers = []
        if domain in self.records:
            records = self.records[domain]
            answers = records.get(rtype, [])
        return Answers(answers, self.source)



"""
Backend Recursive Client-Forwarder Extension
"""
from typing import ClassVar

from pyderive import dataclass

from . import Answers, Backend
from ...client import BaseClient
from ... import RType, Answer, Question

#** Variables **#
__all__ = ['Forwarder']

#** Classes **#

@dataclass(slots=True, repr=False)
class Forwarder(Backend):
    """
    Recursive Dns-Client Lookup Forwarder when Backend returns no Results
    """
    source: ClassVar[str] = 'Forwarder'
    recursion_available: ClassVar[bool] = True #type: ignore

    backend: Backend
    client:  BaseClient

    def is_authority(self, domain: bytes) -> bool:
        return self.backend.is_authority(domain)

    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        """
        query for answers w/ client if base-backend returns empty result
        """
        answers, source = self.backend.get_answers(domain, rtype)
        if not answers:
            source  = self.source
            message = self.client.query(Question(domain, rtype))
            answers.extend(message.answers)
            answers.extend(message.authority)
            answers.extend([
                a for a in message.additional if isinstance(a, Answer)])
        return Answers(answers, source)



"""
EDNS OPT Answer Varient Implementation
"""
from typing import Optional
from typing_extensions import Annotated, Self

from pyderive import dataclass
from pystructs import Context, Struct, Domain, U8, U16

from ..enum import RType
from ..answer import BaseAnswer

#** Variables **#
__all__ = ['ROOT', 'EdnsAnswer']

#: root domain (according to EDNS)
ROOT = b''

#** Classes **#

class EdnsHeader(Struct):
    name:           Domain
    rtype:          Annotated[RType, U16]
    payload_size:   U16
    extended_rcode: U8
    version:        U8
    z:              U16
    data_length:    U16

@dataclass(slots=True)
class EdnsAnswer(BaseAnswer):
    """
    EDNS Answer Object Definition
    """
    name:     bytes  = ROOT
    version:  int    = 0
    content:  bytes  = b''
    udp_size: int    = 512

    @property
    def rtype(self) -> RType:
        return RType.OPT

    def pack(self, ctx: Optional[Context] = None) -> bytes:
        ctx = ctx or Context()
        return EdnsHeader(
            name=self.name,
            rtype=self.rtype,
            payload_size=self.udp_size,
            extended_rcode=0,
            version=self.version,
            z=0,
            data_length=len(self.content)
        ).pack(ctx) + ctx.track_bytes(self.content)

    @classmethod
    def unpack(cls, raw: bytes, ctx: Optional[Context] = None) -> Self:
        ctx     = ctx or Context()
        header  = EdnsHeader.unpack(raw, ctx)
        content = ctx.slice(raw, header.data_length)
        return cls(
            name=header.name,
            version=header.version,
            udp_size=header.payload_size,
            content=content,
        )




"""
DNS Enumeration Definitions
"""
from enum import IntEnum

#** Variables **#
__all__ = [
    'QR',
    'OpCode',
    'RCode',
    'RType',
    'RClass',
    'EDNSOption',
]

#** Classes **#

class QR(IntEnum):
    """
    Message Operation-Code (QUESTION/RESPONSE)
    """
    Question = 0
    Response = 1

class OpCode(IntEnum):
    """
    Message Operation Code - Query, Inverse-Query, etc
    """
    Query        = 0
    InverseQuery = 1
    Status       = 2
    Notify       = 4
    Update       = 5

class RCode(IntEnum):
    """
    Message Response/Error Code
    """
    NoError           = 0
    FormatError       = 1
    ServerFailure     = 2
    NonExistantDomain = 3
    NotImplemented    = 4
    Refused           = 5
    YXDomain          = 6
    YXRRSet           = 7
    NXRRSet           = 8
    NotAuthorized     = 9
    NotInZone         = 10

    BadOPTVersion     = 16
    BadSignature      = 16
    BadKey            = 17
    BadTime           = 18
    BadMode           = 19
    BadName           = 20
    BadAlgorithm      = 21

class RType(IntEnum):
    """
    Question/Action Resource Record (RR) Type
    """
    A          = 1
    NS         = 2
    MD         = 3
    MF         = 4
    CNAME      = 5
    SOA        = 6
    MB         = 7
    MG         = 8
    MR         = 9
    NULL       = 10
    WKS        = 11
    PTR        = 12
    HINFO      = 13
    MINFO      = 14
    MX         = 15
    TXT        = 16
    RP         = 17
    AFSDB      = 18
    X25        = 19
    ISDN       = 20
    RT         = 21
    NSAP       = 22
    NSAP_PTR   = 23
    SIG        = 24
    KEY        = 25
    AAAA       = 28
    LOC        = 29
    NXT        = 30
    EID        = 31
    NB         = 32
    SRV        = 33
    NAPTR      = 35
    KX         = 36
    CERT       = 37
    A6         = 38
    DNAME      = 39
    SINK       = 40
    OPT        = 41
    APL        = 42
    DS         = 43
    SSHFP      = 44
    IPSECKEY   = 46
    RRSIG      = 46
    NSEC       = 47
    DNSKEY     = 48
    DHCID      = 49
    NSEC3      = 50
    NSEC3PARAM = 51
    TLSA       = 52
    SMIMEA     = 53
    HIP        = 55
    NINFO      = 56
    RKEY       = 57
    TALINK     = 58
    CDS        = 59
    CDNSKEY    = 60
    OPENPGPKEY = 61
    CSYNC      = 62
    ZONEMD     = 63
    SVCB       = 64
    HTTPS      = 65

    SPF    = 99
    UINFO  = 100
    UID    = 101
    GID    = 102
    UNPSEC = 103
    NID    = 104
    L32    = 105
    L64    = 106
    LP     = 107
    EUI48  = 108
    EUI64  = 109

    TKEY  = 249
    TSIG  = 250
    AXFR  = 252
    MAILB = 253
    MAILA = 254
    ANY   = 255
    URI   = 256
    CAA   = 257
    DOTA  = 259

    TA  = 32768
    DLV = 32769

class RClass(IntEnum):
    """
    RR Classification (Almost Always IN - Internet)
    """
    IN   = 1
    CS   = 2
    CH   = 3
    HS   = 4
    NONE = 254
    ANY  = 255

class EDNSOption(IntEnum):
    """
    EDNS RR Options
    """
    Cookie = 10



"""
Simple Python DNS Library
"""

#** Variables **#
__all__ = [
    'BaseAnswer',
    'Answer',
    'PreRequisite',
    'Update',

    'Content',
    'Unknown',
    'NULL',
    'ANY',
    'CNAME',
    'MX',
    'NS',
    'PTR',
    'SOA',
    'TXT',
    'A',
    'AAAA',
    'SRV',

    'QR',
    'OpCode',
    'RCode',
    'RType',
    'RClass',
    'EDNSOption',

    'DnsError',
    'ServerFailure',
    'NonExistantDomain',
    'NotImplemented',

    'Flags',
    'Message',

    'Question',
    'Zone'
]

#** Imports **#
from .answer import *
from .content import *
from .enum import *
from .exceptions import *
from .flags import *
from .message import *
from .question import *



"""
DNS RCode Exceptions
"""
from typing import Any, Optional

from .enum import RCode

#** Variables **#
__all__ = [
    'DnsError',
    'ServerFailure',
    'NonExistantDomain',
    'NotImplemented',
]

#** Functions **#

def raise_error(rcode: RCode, message: Any = None):
    """
    raise best exception object to match the given rcode

    :param rcode:   response code from message
    :param message: message to include with exception
    """
    global EXCEPTION_MAP
    eclass = EXCEPTION_MAP.get(rcode, DnsError)
    raise eclass(message, rcode)

#** Classes **#

class DnsError(Exception):
    rcode: RCode = RCode.NoError

    def __init__(self, msg: Any = None, rcode: Optional[RCode] = None):
        self.message = msg
        self.rcode   = rcode or self.rcode

    def __str__(self) -> str:
        if self.message and self.__class__.rcode == self.rcode:
            return str(self.message)
        return super().__str__()

class ServerFailure(DnsError):
    rcode = RCode.ServerFailure

class NonExistantDomain(DnsError):
    rcode = RCode.NonExistantDomain

class NotImplemented(DnsError):
    rcode = RCode.NotImplemented

#** Exceptions **#

#: cheeky way of collecting all exception types into map based on their RCode
EXCEPTION_MAP = {v.rcode:v
    for v in globals().values()
    if isinstance(v, type) and issubclass(v, DnsError) and v is not DnsError}



"""
DNS Message Object Definition
"""
from typing import List, Optional
from typing_extensions import Self

from pyderive import dataclass, field
from pystructs import U16, Context, Struct

from .answer import Answer, BaseAnswer, PreRequisite, Update, peek_rtype
from .edns import EdnsAnswer
from .enum import OpCode, RCode, RType
from .exceptions import raise_error
from .flags import Flags
from .question import Question, Zone

#** Variables **#
__all__ = ['Message']

#** Classes **#

class PacketHeader(Struct):
    id:         U16
    flags:      U16
    questions:  U16
    answers:    U16
    authority:  U16
    additional: U16

@dataclass(slots=True)
class Message:
    """
    DNS Message Object Definition (Request & Response Packet)
    """
    id:         int
    flags:      Flags
    questions:  List[Question]   = field(default_factory=list)
    answers:    List[Answer]     = field(default_factory=list)
    authority:  List[Answer]     = field(default_factory=list)
    additional: List[BaseAnswer] = field(default_factory=list)

    def raise_on_error(self):
        """
        raise exception if message contains an error
        """
        if self.flags.rcode != RCode.NoError:
            domains = list({q.name for q in self.questions})
            domains = domains[0] if len(domains) == 1 else domains
            raise_error(self.flags.rcode, domains or None)

    def pack(self, ctx: Optional[Context] = None) -> bytes:
        """
        pack message object into serialized bytes

        :param ctx: serialization context object
        :return:    serialized bytes
        """
        ctx  = ctx or Context()
        raw  = bytearray()
        raw += PacketHeader(
            id=self.id,
            flags=int(self.flags),
            questions=len(self.questions),
            answers=len(self.answers),
            authority=len(self.authority),
            additional=len(self.additional),
        ).pack(ctx)
        raw += b''.join(q.pack(ctx) for q in self.questions)
        raw += b''.join(a.pack(ctx) for a in self.answers)
        raw += b''.join(a.pack(ctx) for a in self.authority)
        raw += b''.join(a.pack(ctx) for a in self.additional)
        return bytes(raw)

    @classmethod
    def unpack(cls, raw: bytes, ctx: Optional[Context] = None) -> Self:
        """
        unpack serialized bytes into deserialized message object

        :param raw: raw byte buffer
        :param ctx: deserialization context object
        :return:    unpacked message object
        """
        ctx   = ctx or Context()
        head  = PacketHeader.unpack(raw, ctx)
        flags = Flags.fromint(head.flags)
        # determine classes to parse content
        qclass, anclass, auclass = (Question, Answer, Answer) \
            if flags.op != OpCode.Update else \
            (Zone, PreRequisite, Update)
        # parse body content w/ determined classes
        questions  = [qclass.unpack(raw, ctx) for _ in range(head.questions)]
        answers    = [anclass.unpack(raw, ctx) for _ in range(head.answers)]
        authority  = [auclass.unpack(raw, ctx) for _ in range(head.authority)]
        additional = []
        for _ in range(head.additional):
            rtype  = peek_rtype(raw, ctx)
            newcls = EdnsAnswer if rtype == RType.OPT else cls
            answer = newcls.unpack(raw, ctx)
            additional.append(answer)
        return cls(
            id=head.id,
            flags=flags,
            questions=questions,
            answers=answers,
            authority=authority,
            additional=additional
        )



"""
Simple DBM Database Implementaton for Rule Engine
"""
import re
import os
import dbm
import json
from typing import List, Optional, Set, Tuple

from pydns.server.backend.ruleset.wildcard import WildcardMatch

from . import RuleEngine
from .parser import RuleDef, RuleDefs, Domain, Regex, Status, Wildcard, parse_rules
from .wildcard import WildcardMatch

#** Variables **#
__all__ = ['DbmRuleEngine']

RegexRules = List[Tuple[re.Pattern, Status]]

WildcardRules = List[Tuple[WildcardMatch, Status]]

#** Functions **#

def open_dbm(path: str, flags: str):
    """
    retrieve dbm object and ensure not dbm.dumb

    :param path:  filepath to dbm database
    :param flags: flags to use when loading database
    """
    database = dbm.open(path, flag=flags) #type: ignore
    which    = dbm.whichdb(path)
    if which is None or which == 'dbm.dumb':
        raise RuntimeError('Python has no valid DBM engine installed.')
    return database

def encode_defs(rules: RuleDefs) -> bytes:
    """
    serialize rule definitions as bytes

    :param rules: iterator of rules to serialize
    :return:      serialized rules
    """
    return json.dumps([ruledef.asdict() for ruledef in rules]).encode('latin1')

def decode_defs(ruledefs: Optional[bytes]) -> List[RuleDef]:
    """
    deserialize rule defintions from bytes

    :param ruledefs: serialized rule definitions
    :return:         deserialized rule definitions
    """
    if not ruledefs:
        return []
    return [RuleDef(**rdef) for rdef in json.loads(ruledefs.decode('latin1'))]

#** Classes **#

class DbmRuleEngine(RuleEngine):
    __slots__ = ('dbm', 'regex', 'wildcards')

    source_key:   str = '__sources'
    regex_key:    str = '__%s_regex'
    wildcard_key: str = '__%s_wildcards'
    domain_key:   str = '__%s_domains'

    def _reload_patterns(self):
        """compile regex and wildcard expressions within database"""
        self.regex     = []
        self.wildcards = []
        for source in self.sources():
            regex_key = self.regex_key % source
            for rdef in decode_defs(self.dbm.get(regex_key)):
                rgx = (re.compile(rdef.rule.encode()), rdef.status)
                self.regex.append(rgx)
            wildcard_key = self.wildcard_key %  source
            for rdef in decode_defs(self.dbm.get(wildcard_key)):
                wild = (WildcardMatch.compile(rdef.rule), rdef.status)
                self.wildcards.append(wild)

    def __init__(self, path: str, flag = 'c'):
        self.dbm = open_dbm(path, flag)
        self._reload_patterns()

    def sources(self) -> Set[str]:
        """
        retrieve list of ingested sources
        """
        return set(self.dbm.get(self.source_key, b'').decode().split(','))

    def ingest(self, name: str, rules: RuleDefs):
        """
        ingest incoming source of rule definitions and update database

        :param name:  name of source
        :param rules: rules to ingest
        """
        # separate rules into categories
        domains   = {}
        regex     = []
        wildcards = []
        for ruledef in rules:
            if isinstance(ruledef.rule, Domain):
                domains[ruledef.rule] = 'b' if ruledef.status else 'w'
            elif isinstance(ruledef.rule, Wildcard):
                wildcards.append(ruledef)
            elif isinstance(ruledef.rule, Regex):
                regex.append(ruledef)
        # retrieve related source keys
        regex_key    = self.regex_key % name
        wildcard_key = self.wildcard_key % name
        domain_key   = self.domain_key % name
        # delete previous domain records (if they exist)
        for domain in self.dbm.get(domain_key, b'').split(b','):
            if domain in self.dbm:
                del self.dbm[domain]
        # update sources
        sources = self.sources()
        sources.add(name)
        # write content into dbm at once
        self.dbm[self.source_key] = ','.join(sources).encode()
        self.dbm[domain_key] = b','.join(d.encode() for d in domains.keys())
        self.dbm[regex_key] = encode_defs(regex)
        self.dbm[wildcard_key] = encode_defs(wildcards)
        for domain, rule in domains.items():
            self.dbm[domain] = rule
        # sync and reorganize data (if available)
        if hasattr(self.dbm, 'sync'):
            self.dbm.sync() #type: ignore
        if hasattr(self.dbm, 'reorganize'):
            self.dbm.reorganize() #type: ignore
        # reload regex/wildcard expressions
        self._reload_patterns()

    def ingest_file(self, fpath: str, name: Optional[str] = None):
        """
        parse and ingest ruleset from the specified filepath

        :param fpath: filepath containing rules to add to rule engine
        :param name:  custom name of source for items in db
        """
        # only ingest the file if it hasnt been seen before or mtime changed
        name = name or os.path.basename(fpath)
        time = os.path.getmtime(fpath)
        last = float(self.dbm.get(fpath, b'0').decode())
        if time == last:
            return
        # process file and ingest domains and then cache last mtime
        with open(fpath, 'r') as f:
            rules = parse_rules(f)
            self.ingest(name, rules)
            self.dbm[fpath] = str(time).encode()

    def match_domain(self, domain: bytes) -> Optional[bool]:
        """
        match domain against dbm database of rules

        :param domain: domain to check if in database
        :return:       rule determination (if found)
        """
        rule = self.dbm.get(domain)
        return rule == b'b' if rule is not None else None

    def match_pattern(self, domain: bytes) -> Optional[bool]:
        """
        match domain against existing pattern based rules

        :param domain: domain to check if matching pattern rules
        :return:       rule determination (if matched)
        """
        for wildcard, rule in self.wildcards:
            if wildcard.match(domain):
                return rule
        for regex, rule in self.regex:
            if regex.match(domain):
                return rule



"""
DNS Answer RR Content Definitions
"""
from functools import lru_cache
from typing import ClassVar, Optional, Type
from typing_extensions import Annotated, Self

from pystructs import U16, U32, Context, Domain, HintedBytes, IPv4, IPv6, Struct

from . enum import RType

#** Variables **#
__all__ = [
    'Content',
    'Unknown',

    'NULL',
    'ANY',
    'CNAME',
    'MX',
    'NS',
    'PTR',
    'SOA',
    'TXT',
    'A',
    'AAAA',
    'SRV',
]

#** Classes **#

class Content(Struct):
    """
    Abstract Baseclass for DNS RR Record Content
    """
    rtype: ClassVar[RType]

class NULL(Content):
    """
    NULL RR - Signifies that query does not exist
    """
    rtype: ClassVar[RType] = RType.NULL

class ANY(Content):
    """
    ANY RR - Deprecated special record type
    """
    rtype: ClassVar[RType] = RType.ANY

class CNAME(Content):
    """
    CNAME RR - Cannoical Name (domain alias)
    """
    rtype: ClassVar[RType] = RType.CNAME
    name:  Domain

class MX(Content):
    """
    MX RR - Mail Server
    """
    rtype:      ClassVar[RType] = RType.MX
    preference: U16
    exchange:   Domain

class NS(Content):
    """
    NS RR - Name Server (authoritative dns zone for domain)
    """
    rtype:      ClassVar[RType] = RType.NS
    nameserver: Domain

class PTR(Content):
    """
    PTR RR - Pointer (reverse ip to domain lookup)
    """
    rtype:   ClassVar[RType] = RType.PTR
    ptrname: Domain

class SOA(Content):
    """
    SOA RR - Start of Authority (authoritative info for domain)
    """
    rtype:     ClassVar[RType] = RType.SOA
    mname:     Domain
    rname:     Domain
    serialver: U32
    refresh:   U32
    retry:     U32
    expire:    U32
    minimum:   U32

class TXT(Content):
    """
    TXT RR - Text Record (Arbitrary Text Blob over DNS)
    """
    rtype: ClassVar[RType] = RType.TXT
    text:  Annotated[bytes, HintedBytes(U32)]

class A(Content):
    """
    A RR - Address Record (ipv4 address)
    """
    rtype: ClassVar[RType] = RType.A
    ip:    IPv4

class AAAA(Content):
    """
    AAAA RR - Ipv6 Address Record (ipv6 address)
    """
    rtype: ClassVar[RType] = RType.AAAA
    ip:    IPv6

class SRV(Content):
    """
    SRV RR - Service Record (generalized service rather than NS/MX)
    """
    rtype:    ClassVar[RType] = RType.SRV
    priority: U16
    weight:   U16
    port:     U16
    target:   Domain

class Unknown:
    """
    Mock Struct/Content Object for Unknown/Unsupported DNS Content Types
    """
    __slots__ = ('data', )

    rtype: ClassVar[RType]
    size:  ClassVar[int]

    def __init__(self, data: bytes):
        self.data = data

    def __repr__(self) -> str:
        return f'Unknown(rtype={self.rtype!r}, data=0x{self.data.hex()})'

    def pack(self, ctx: Optional[Context] = None) -> bytes:
        ctx = ctx or Context()
        return ctx.track_bytes(self.data)

    @classmethod
    def unpack(cls, raw: bytes, ctx: Optional[Context] = None) -> Self:
        ctx = ctx or Context()
        return cls(ctx.slice(raw, cls.size))

    @classmethod
    @lru_cache(maxsize=None)
    def new(cls, rtype: RType, size: int) -> Type:
        return type('Unknown', (cls, ), {'rtype': rtype, 'size': size})

#** Content **#

#: cheeky way of collecting all content types into map based on their RType
CONTENT_MAP = {v.rtype:v
    for v in globals().values()
    if isinstance(v, type) and issubclass(v, Content) and v is not Content}



"""
RulesList Parser AdGuard/Domain-List/uBlock/etc...
"""
import re
import ipaddress
import warnings
from typing import Iterable, NewType, Optional, TextIO, Union

from pyderive.extensions.serde import Serde

#** Variables **#
__all__ = [
    'Status',
    'Domain',
    'Regex',
    'Rule',
    'RuleDef',
    'RuleDefs',

    'parse_rule',
    'parse_rules',
]

#: wrapper around whitelist/blacklist determination
Status = NewType('Status', bool)

#: wrapper around domain string
class Domain(str): pass

#: wrapper around regex string
class Regex(str): pass

#: wrapper around wildcard string
class Wildcard(str): pass

#: Possible Rules (Domain/Wildcard/Regex)
Rule = Union[Domain, Regex, Wildcard]

class RuleDef(Serde, slots=True):
    rule:   Rule
    status: Status

#: generator of rule results
RuleDefs = Iterable[RuleDef]

#: raw regex expression to match valid web domains
re_expr = r'(?:[a-zA-Z0-9_](?:[a-zA-Z0-9-_]{0,61}' + \
    r'[A-Za-z0-9])?\.)+[A-Za-z0-9][A-Za-z0-9-_]{0,61}[A-Za-z]\.?'

#: compiled regex expression used to find domains in string
domain_find = re.compile(re_expr, re.IGNORECASE)

#: compiled regex expression to match domains as full-string only
domain_exact = re.compile(f'^{re_expr}$', re.IGNORECASE)

#: allowed (but not supported) adguard options
ALLOWED_OPTIONS = {'dnsrewrite', 'important'}

#** Functions **#

#NOTE: current implementation ignores existing adguard filter
# options which can effect how rules are applied.
# (https://adguard-dns.io/kb/general/dns-filtering-syntax)

def is_ipaddr(ip: str) -> bool:
    """
    attempt to parse string as an ip-address and return success

    :param ip: potential ip-address string
    :return:   true if string is ip-address
    """
    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False

def is_regex(rgx: str) -> bool:
    """
    attempt to parse string as regex expression and return success

    :param rgx: potential regex string
    :return:    true if string is valid regex
    """
    try:
        re.compile(rgx)
        return True
    except re.error:
        return False

def parse_rule(line: str) -> Optional[RuleDef]:
    """
    parse rule string into associated rule-type

    :param line: raw rule defintion on single-line
    :return:     parsed rule definition
    """
    # handle regex expr rule
    status = Status(not line.startswith('@@'))
    if line.startswith('/'):
        line = line.strip('/')
        if not is_regex(line):
            return
        return RuleDef(Regex(line), status)
    # parse adguard options from rule: ||example.com^$settings=1
    rule = line.strip('@')
    if rule.startswith('|') and '$' in rule:
        rule, options = rule.rsplit('$', 1)
        if not any(opt in options for opt in ALLOWED_OPTIONS):
            return
    # check if rule is in old `/etc/hosts` style format: `0.0.0.0 example.com`
    if ' ' in rule:
        start, end = rule.split(' ', 1)
        if is_ipaddr(start):
            rule = end.split('#', 1)[0].strip()
    # strip url extensions from rules
    for c in '/#?':
        if c in rule:
            rule = rule.split(c, 1)[0].strip()
    # handle traditional domain rules: ||www.google.com^
    # by adguard standards this is still a wildcard without ending with a `|`
    # but processing it like a flat domain comparison is better and faster
    # and acts exactly the same when its just a domain
    clean_rule = rule.strip('|^')
    if domain_exact.match(clean_rule):
       return RuleDef(Domain(clean_rule), status)
    # ignore ip-address rules since they won't apply to anything: ||1.2.3.4^
    if is_ipaddr(clean_rule):
        return
    # treat non-domains without a clear ending as a wildcard: ||abcdef.^
    if not any(rule.startswith(c) for c in '|*'):
        clean_rule = '*' + rule
    if not any(rule.endswith(c) for c in '^|*'):
        clean_rule += '*'
    # handle wildcard domain rules. anything with a `*` is safely a wildcard
    if '*' in clean_rule:
        return RuleDef(Wildcard(clean_rule), status)
    # error on failure to parse
    warnings.warn(f'Invalid Rule: {line!r}')

def parse_rules(f: TextIO) -> RuleDefs:
    """
    parse all available rules contained within a file

    :param f: file object to read rules from
    :return:  generator of parsed rule definitions
    """
    lines = (l.strip() for l in f.read().split('\n'))
    lines = (l for l in lines if l)
    for line in lines:
        if any(line.startswith(c) for c in ('!', '#', ':')):
            continue
        rule = parse_rule(line)
        if rule is not None:
            yield rule



"""
DNS Client UnitTests
"""
from unittest import TestCase

from .. import Question, RCode, RType
from ..client import HttpsClient, TcpClient, UdpClient

#** Variables **#
__all__ = ['ClientTests']

#** Classes **#

class ClientTests(TestCase):
    """
    DNS Message Packet Parsing/Construction UnitTests
    """

    def test_udp_client(self):
        """
        ensure udp client works as intended
        """
        client   = UdpClient([('1.1.1.1', 53)])
        response = client.query(Question(b'one.one.one.one', RType.A))
        self.assertEqual(response.flags.rcode, RCode.NoError)
        self.assertEqual(len(response.answers), 2)
        self.assertEqual(response.answers[0].rtype, RType.A)
        self.assertEqual(response.answers[1].rtype, RType.A)
        self.assertEqual({str(a.content.ip) for a in response.answers}, #type: ignore
            {'1.0.0.1', '1.1.1.1'})

    def test_tcp_client(self):
        """
        ensure tcp client works as intended
        """
        client   = TcpClient([('1.1.1.1', 53)])
        response = client.query(Question(b'one.one.one.one', RType.A))
        self.assertEqual(response.flags.rcode, RCode.NoError)
        self.assertEqual(len(response.answers), 2)
        self.assertEqual(response.answers[0].rtype, RType.A)
        self.assertEqual(response.answers[1].rtype, RType.A)
        self.assertEqual({str(a.content.ip) for a in response.answers}, #type: ignore
            {'1.0.0.1', '1.1.1.1'})

    def test_https_client(self):
        """
        ensure https client works as intended
        """
        client   = HttpsClient()
        response = client.query(Question(b'one.one.one.one', RType.A))
        self.assertEqual(response.flags.rcode, RCode.NoError)
        self.assertEqual(len(response.answers), 2)
        self.assertEqual(response.answers[0].rtype, RType.A)
        self.assertEqual(response.answers[1].rtype, RType.A)
        self.assertEqual({str(a.content.ip) for a in response.answers}, #type: ignore
            {'1.0.0.1', '1.1.1.1'})



"""
Standard UDP/TCP Client Implementations
"""
import random
import socket
from abc import ABC, abstractmethod
from typing import List, Optional

from pypool import Pool
from pyserve import RawAddr
from pyderive import dataclass

from . import BaseClient, Message

#** Variables **#
__all__ = ['UdpClient', 'TcpClient']

#** Classes **#

class SocketPool(Pool[socket.socket]):
    pass

@dataclass(slots=True)
class Client(BaseClient, ABC):
    """
    Baseclass Socket-Based DNS Client Implementation
    """
    addresses:  List[RawAddr]
    block_size: int           = 65535
    pool_size:  Optional[int] = None
    expiration: Optional[int] = 15
    timeout:    int           = 10

    def __post_init__(self):
        self.pool = SocketPool(
            factory=self.newsock,
            cleanup=self.cleanup,
            max_size=self.pool_size,
            expiration=self.expiration)

    @abstractmethod
    def newsock(self) -> socket.socket:
        """
        spawn new socket object to make request

        :return: new socket object
        """
        raise NotImplementedError

    @abstractmethod
    def cleanup(self, sock: socket.socket):
        """
        close and clean an existing socket object

        :param sock: socket object
        """
        raise NotImplementedError

    def pickaddr(self) -> RawAddr:
        """
        pick random address from list of addresses

        :return: random dns address to make request
        """
        return random.choice(self.addresses)

    def drain(self):
        """
        drain socket pool
        """
        self.pool.drain()

class UdpClient(Client):
    """
    Simple UDP Socket DNS Client
    """

    def newsock(self) -> socket.socket:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(self.timeout)
        return sock

    def cleanup(self, sock: socket.socket):
        sock.close()

    def request(self, msg: Message) -> Message:
        with self.pool.reserve() as sock:
            # send request
            addr = self.pickaddr()
            data = msg.pack()
            sock.sendto(data, addr)
            # recieve response
            data, _ = sock.recvfrom(self.block_size)
            return Message.unpack(data)

class TcpClient(Client):
    """
    Simple TCP Socket DNS Client
    """

    def newsock(self) -> socket.socket:
        addr = self.pickaddr()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        sock.connect(addr)
        return sock

    def cleanup(self, sock: socket.socket):
        sock.shutdown(socket.SHUT_RDWR)
        sock.close()

    def request(self, msg: Message) -> Message:
        with self.pool.reserve() as sock:
            # send request
            data = msg.pack()
            data = len(data).to_bytes(2, 'big') + data
            sock.send(data)
            # recieve size of response
            sizeb = sock.recv(2)
            size  = int.from_bytes(sizeb, 'big')
            # read data from size
            data = sock.recv(size)
            return Message.unpack(data)




"""
DNS Answer Object Definitions
"""
from abc import abstractmethod
from typing import Optional, Protocol, Type
from typing_extensions import Annotated, Self

from pyderive import dataclass, field
from pystructs import U16, U32, Context, Domain, Struct

from .enum import RType, RClass
from .content import ANY, CONTENT_MAP, Content, Unknown

#** Variables **#
__all__ = ['BaseAnswer', 'Answer', 'PreRequisite', 'Update']

#** Functions **#

def get_ctype(rtype: RType, size: Optional[int] = None) -> Type[Content]:
    """
    retrieve record content type based on record-type (and size if specified)

    :param rtype: record-type of requested object type
    :param size:  optional size-hint for record-type
    :return:      content record type
    """
    if rtype in CONTENT_MAP:
        return CONTENT_MAP[rtype]
    if size is not None:
        return Unknown.new(rtype, size)
    raise ValueError(f'Unsupported RR Type: {rtype.name!r}')

def peek_rtype(raw: bytes, ctx: Context) -> RType:
    """
    peek next anwser in buffer to retrieve specified RR Type
    """
    idx  = ctx.index
    peek = PeekHeader.unpack(raw, ctx)
    ctx.index = idx
    return peek.rtype

#** Classes **#

#NOTE: The nature of dns domain name compression prevents
# the simple implementation of including the content-size
# and content-body within `AnswerHeader` with something like
# a `HintedBytes(U16)` instance which would require recursively
# packing/unpacking the content assigned to a bytearray.

class PeekHeader(Struct):
    name:  Domain
    rtype: Annotated[RType, U16]

class AnswerHeader(PeekHeader):
    rclass:  Annotated[RClass, U16]
    ttl:     U32

class BaseAnswer(Protocol):
    """
    Baseclass for defining `Answer` objects
    """
    name: bytes

    @property
    @abstractmethod
    def rtype(self) -> RType:
        raise NotImplementedError

    @abstractmethod
    def pack(self, ctx: Optional[Context] = None) -> bytes:
        """
        pack answer object into serialized bytes

        :param ctx: serialization context object
        :return:    serialized bytes
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def unpack(cls, raw: bytes, ctx: Optional[Context] = None) -> Self:
        """
        unpack serialized bytes into deserialized answer object

        :param raw: raw byte buffer
        :param ctx: deserialization context object
        :return:    unpacked answer object
        """
        raise NotImplementedError

@dataclass(slots=True)
class Answer(BaseAnswer):
    """
    Standard DNS Answer Implementation
    """
    name:    bytes
    ttl:     int
    content: Content
    rclass:  RClass = RClass.IN

    @property
    def rtype(self) -> RType:
        return self.content.rtype

    def pack(self, ctx: Optional[Context] = None) -> bytes:
        ctx  = ctx or Context()
        head = AnswerHeader(name=self.name, rtype=self.content.rtype,
            rclass=self.rclass, ttl=self.ttl).pack(ctx)
        ctx.index += 2 # pre-increment index (placeholder for size)
        body = self.content.pack(ctx)
        size = len(body).to_bytes(2, 'big')
        return head + size + body

    @classmethod
    def unpack(cls, raw: bytes, ctx: Optional[Context] = None) -> Self:
        ctx     = ctx or Context()
        header  = AnswerHeader.unpack(raw, ctx)
        size    = int.from_bytes(ctx.slice(raw, 2), 'big')
        ctype   = get_ctype(header.rtype, size)
        content = ctype.unpack(raw, ctx)
        return cls(header.name, header.ttl, content, header.rclass)

@dataclass(slots=True)
class PreRequisite(Answer):
    """
    Alias for Answer in UPDATE action DNS Requests with Sensible Defaults
    """
    ttl:     int = 0
    content: Content = field(default=ANY)

class Update(Answer):
    """
    Alias for Answer in UPDATE action DNS Requests
    """
    pass



"""
Custom Rule Engine for Blocking Unwanted Domains
"""
from abc import abstractmethod
from typing import ClassVar, List, Optional, Protocol, Set

from pyderive import dataclass, field

from .. import RType, Answers, Backend

#** Variables **#
__all__ = [
    'RuleEngine',
    'RuleBackend',

    'DbmRuleEngine',
]

#** Functions **#

def split_domains(domain: bytes) -> List[bytes]:
    """
    split base-domain into all possible subdomains
    example: www.example.com => [www.example.com, example.com]

    :param domain: domain to be split into items
    :return:       split domain list
    """
    domains = []
    while domain.count(b'.') > 0:
        domains.append(domain)
        _, domain = domain.split(b'.', 1)
    return domains

#** Classes **#

class RuleEngine(Protocol):
    """
    Abstract Rule Engine Implementation for Domain Blocking
    """

    @abstractmethod
    def match_domain(self, domain: bytes) -> Optional[bool]:
        raise NotImplementedError

    @abstractmethod
    def match_pattern(self, domain: bytes) -> Optional[bool]:
        raise NotImplementedError

    def match(self, domain: bytes) -> Optional[bool]:
        """
        check if domain is contained within the database

        :param domain: domain to check if in db
        :return:       true if domain in db
        """
        for match in split_domains(domain):
            rule = self.match_domain(match)
            if rule is not None:
                return rule
        return self.match_pattern(domain)

@dataclass(slots=True, repr=False)
class RuleBackend(Backend):
    """
    Custom Rule Engine Backend for Blacklisting Unwanted Domains
    """
    source: ClassVar[str] = 'Blacklist'

    backend:   Backend
    blacklist: Set[bytes]           = field(default_factory=set)
    whitelist: Set[bytes]           = field(default_factory=set)
    engine:    Optional[RuleEngine] = None

    def __post_init__(self):
        self.recursion_available = self.backend.recursion_available
        self.empty = Answers([], self.source)
        self.blacklist -= self.whitelist

    def is_authority(self, domain: bytes) -> bool:
        """
        :param domain: check if domain is authority
        :return:       true if backend has authority over domain
        """
        return self.backend.is_authority(domain)

    def is_blocked(self, domain: bytes) -> bool:
        """
        check if the following domain is blocked

        :param domain: domain to check if blocked
        :return:       true if domain is blocked else false
        """
        # check if domain in in-memory whitelist of blacklist
        domains = split_domains(domain)
        if any(match in self.whitelist for match in domains):
            return False
        if any(match in self.blacklist for match in domains):
            return True
        # check if record in database
        if self.engine:
            # match against exact domain
            for match in domains:
                matches = self.engine.match_domain(match)
                if matches is True:
                    self.blacklist.add(domain)
                    self.blacklist.add(match)
                    return True
                if matches is False:
                    self.whitelist.add(domain)
                    self.whitelist.add(match)
                    return False
            # match against patterns
            matches = self.engine.match_pattern(domain)
            if matches is True:
                self.blacklist.add(domain)
                return True
            if matches is False:
                self.whitelist.add(domain)
                return False
        return False

    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        """
        block lookups for blacklisted domains, otherwise do standard query

        :param domain: domain to check if blocked or return results
        :param rtype:  record-type associated w/ query
        :return:       empty-answers (if blocked), else standard search results
        """
        if self.is_blocked(domain):
            return self.empty
        return self.backend.get_answers(domain, rtype)

#** Imports **#
from .database import DbmRuleEngine
